/*
  File name: CCytadelaGame.cpp
  Copyright: (C) 2005 - 2007 Tomasz Kazmierczak

  Creation date: 18.07.2005 15:37
  Last modification date: 8.09.2007

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

  game main loop and some other functions
*/

#include <iostream>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <memory.h>
#include "CCytadelaGame.h"
#include "cmf/CMFile.h"

using namespace std;

enum {
	OSDIMG_PANEL = 0,
	OSDIMG_WEAPON_HAND,
	OSDIMG_WEAPON_HG,
	OSDIMG_WEAPON_SG,
	OSDIMG_WEAPON_MG,
	OSDIMG_WEAPON_FT,
	OSDIMG_WEAPON_BL,
	OSDIMG_WEAPON_RL,
	OSDIMG_WEAPON_RC,
	OSDIMG_WEAPON_GC,
	OSDIMG_WEAPON_BC,
	OSDIMG_LDIGIT_NA,
	OSDIMG_LDIGIT_0,
	OSDIMG_SDIGIT_0 = 22,
	OSDIMG_RDIGIT_0 = 32,
	OSDIMG_BOMB_0 = 42,
	OSDIMG_PLOTTER = 49,
	OSDIMG_OSDFONT,

	OSDIMG_IMAGES_COUNT
};

bool CCytadelaGame::initGame(COGLClass *pOGL, GameState *pState, CInfoText *pTexts,
							 CCytadelaConfig *pConfig)
{
	//copy pointers to objects
	mSDL = CSDLClass::instance();
	mOGL = pOGL;
	mTexts = pTexts;
	mLocInterface = ILocalization::instance();
	mConfig = pConfig;

	//initialize global variables
	mRotY = mSin = mCos = mX = mZ = 0.0f;
	mEnergy = pState->wEnergy;
	mBomb = pState->btBomb;
	mDifficult = pState->bDifficult;
	mKilled = 0;
	mPause = false;
	mFireOnce = false;
	mMap = false;
	mTotalTime = 0;
	mLastErrorDesc = "NO ERROR";

	//a buffer to store the standard walk speed - it shouldn't be changed
	//during the gameplay
	//(in the original Cytadela it was an integer equal to 250 - 1/4 field per
	//one frame)
	//multiply it by game speed
	mStandardWalkSpeed = 0.1f * mConfig->getGameSpeed();
	//at the begining the player walks with the standard speed
	mWalkSpeed = mStandardWalkSpeed;
	mPotion = 0;
	mBeer = 0;
	mDiff = 0;

	mFinishedLevel = false;

	for(uint32_t i = 0; i < (uint32_t)ACT_NUM_ACTIONS; ++i)
		mActionMap[i] = false;

	//init hand object
	mUsableItems[UIT_HAND] = (UsableItem *)(new Hand);
	//select the hand
	mSelectedItem = mUsableItems[UIT_HAND];
	//weapons
	mUsableItems[UIT_HANDGUN] = (UsableItem *)(new Weapon(UIT_HANDGUN, 1, 5, 800.0f, Ammo(10)));
	mUsableItems[UIT_SHOTGUN] = (UsableItem *)(new Weapon(UIT_SHOTGUN, 3, 5, 1000.0f, Ammo(5)));
	mUsableItems[UIT_MASHINEGUN] = (UsableItem *)(new Weapon(UIT_MASHINEGUN, 2, 5, 1000.0f, Ammo(10)));
	mUsableItems[UIT_FLAMER] = (UsableItem *)(new Weapon(UIT_FLAMER, 1, 25, 1200.0f, Ammo(5)));
	mUsableItems[UIT_BLASTER] = (UsableItem *)(new Weapon(UIT_BLASTER, 1, 20, 1200.0f, Ammo(5)));
	mUsableItems[UIT_ROCKETLAUNCHER] = (UsableItem *)(new Weapon(UIT_ROCKETLAUNCHER, 1, 40, 1200.0f, Ammo(1)));
	//cards
	mUsableItems[UIT_RED_CARD] = (UsableItem *)(new Card(UIT_RED_CARD));
	mUsableItems[UIT_GREEN_CARD] = (UsableItem *)(new Card(UIT_GREEN_CARD));
	mUsableItems[UIT_BLUE_CARD] = (UsableItem *)(new Card(UIT_BLUE_CARD));
	//check if the objects were created
	for(UsableItemsMap::iterator i = mUsableItems.begin(); i != mUsableItems.end(); i++)
	{
		if(not i->second)
		{
			fprintf(stderr, "CCytadelaGame::initGame(): !mUsableItems[i] - Out of memory\n");
			mLastErrorDesc = "ERROR: OUT OF MEMORY!";
			return false;
		}
	}

	//set the state of the weapons
	//handgun
	Weapon *pWeapon = (Weapon *)mUsableItems[UIT_HANDGUN];
	if(pState->wWeapons & GSWPN_HANDGUN)
		pWeapon->setStatus(WPNS_ACTIVE);
	pWeapon->getAmmo()->setAmount(pState->wAmmo[GSAM_HANDGUN]);
	//shotgun
	pWeapon = (Weapon *)mUsableItems[UIT_SHOTGUN];
	if(pState->wWeapons & GSWPN_SHOTGUN)
		pWeapon->setStatus(WPNS_ACTIVE);
	pWeapon->getAmmo()->setAmount(pState->wAmmo[GSAM_SHOTGUN]);
	//mashinegun
	pWeapon = (Weapon *)mUsableItems[UIT_MASHINEGUN];
	if(pState->wWeapons & GSWPN_MASHINEGUN)
		pWeapon->setStatus(WPNS_ACTIVE);
	pWeapon->getAmmo()->setAmount(pState->wAmmo[GSAM_MASHINEGUN]);
	//flamer
	pWeapon = (Weapon *)mUsableItems[UIT_FLAMER];
	if(pState->wWeapons & GSWPN_FLAMER)
		pWeapon->setStatus(WPNS_ACTIVE);
	pWeapon->getAmmo()->setAmount(pState->wAmmo[GSAM_FLAMER]);
	//blaster
	pWeapon = (Weapon *)mUsableItems[UIT_BLASTER];
	if(pState->wWeapons & GSWPN_BLASTER)
		pWeapon->setStatus(WPNS_ACTIVE);
	pWeapon->getAmmo()->setAmount(pState->wAmmo[GSAM_BLASTER]);
	//launcher
	pWeapon = (Weapon *)mUsableItems[UIT_ROCKETLAUNCHER];
	if(pState->wWeapons & GSWPN_ROCKETLAUNCHER)
		pWeapon->setStatus(WPNS_ACTIVE);
	pWeapon->getAmmo()->setAmount(pState->wAmmo[GSAM_ROCKETLAUNCHER]);

	//simple items
	mSimpleItems[SIT_MEDIPACK] = new Item(30);
	mSimpleItems[SIT_POTION] = new Item(75000); //300 * 250ms
	mSimpleItems[SIT_BEER] = new Item(47500); //190 * 250ms
	mSimpleItems[SIT_BOMB] = new Item(1);
	//check if the objects were created
	for(SimpleItemsMap::iterator i = mSimpleItems.begin(); i != mSimpleItems.end(); i++)
	{
		if(not i->second)
		{
			fprintf(stderr, "CCytadelaGame::initGame(): !mSimpleItems[i] - Out of memory\n");
			mLastErrorDesc = "ERROR: OUT OF MEMORY!";
			return false;
		}
	}

	//run the random numbers' generator
	srand(time(0));
	return true;
}

void CCytadelaGame::freeGame()
{
	for(SimpleItemsMap::iterator i = mSimpleItems.begin(); i != mSimpleItems.end(); i++)
	{
		if(i->second)
			delete i->second;
	}
	mSimpleItems.clear();
	for(UsableItemsMap::iterator i = mUsableItems.begin(); i != mUsableItems.end(); i++)
	{
		if(i->second)
			delete i->second;
	}
	mUsableItems.clear();

	mSDL->freeSounds();
	mSDL = 0;
	mOGL = 0;
}

const char *CCytadelaGame::getLastErrorDesc()
{
	return mLastErrorDesc;
}

bool CCytadelaGame::loadImages()
{
	//image names
	char *szImgNames[] = {"Okno.tga", "e01.tga", "e02.tga",
		"e03.tga", "e04.tga", "e05.tga", "e06.tga", "e07.tga", "e14.tga",
		"e15.tga", "e16.tga", "ldn.tga", "ld0.tga", "ld1.tga", "ld2.tga",
		"ld3.tga", "ld4.tga", "ld5.tga", "ld6.tga", "ld7.tga", "ld8.tga",
		"ld9.tga", "sd0.tga", "sd1.tga", "sd2.tga", "sd3.tga", "sd4.tga",
		"sd5.tga", "sd6.tga", "sd7.tga", "sd8.tga", "sd9.tga", "rd0.tga",
		"rd1.tga", "rd2.tga", "rd3.tga", "rd4.tga", "rd5.tga", "rd6.tga",
		"rd7.tga", "rd8.tga", "rd9.tga", "b0.tga", "b1.tga", "b2.tga",
		"b3.tga", "b4.tga", "b5.tga", "b6.tga", "plot.tga", "gamefont.tga"};
	//image paths
	char *szImgPaths[OSDIMG_IMAGES_COUNT];
	//the same for (almost) all
	for(int32_t i = 0; i < OSDIMG_IMAGES_COUNT; ++i)
		szImgPaths[i] = "tex/Gfx.far";
	//font path depends on the chosen localization
	char szFontPath[FILENAME_MAX];
	sprintf(szFontPath, "tex/fonts/%s.far", mLocInterface->getCurrentLocalization());
	szImgPaths[OSDIMG_OSDFONT] = szFontPath;

	//create an images array
	if(not mOGL->mOSD.createImagesArray(OSDIMG_IMAGES_COUNT))
		return false;
	//load the images
	for(uint32_t i = 0; i < OSDIMG_IMAGES_COUNT; i++)
		if(not mOGL->mOSD.loadImg(i, szImgPaths[i], szImgNames[i]))
			return false;

	return true;
}

bool CCytadelaGame::loadItemSounds()
{
	int32_t iSound = mSDL->loadSound("sounds/handgun.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_HANDGUN]->mUseSoundNumber = iSound;
	iSound = mSDL->loadSound("sounds/shotgun.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_SHOTGUN]->mUseSoundNumber = iSound;
	iSound = mSDL->loadSound("sounds/machine2.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_MASHINEGUN]->mUseSoundNumber = iSound;
	iSound = mSDL->loadSound("sounds/flamer.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_FLAMER]->mUseSoundNumber = iSound;
	iSound = mSDL->loadSound("sounds/blaster.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_BLASTER]->mUseSoundNumber = iSound;
	iSound = mSDL->loadSound("sounds/launcher.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_ROCKETLAUNCHER]->mUseSoundNumber = iSound;

	iSound = mSDL->loadSound("sounds/noammo.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	((Weapon *)(mUsableItems[UIT_HANDGUN]))->mNoAmmoSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_SHOTGUN]))->mNoAmmoSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_MASHINEGUN]))->mNoAmmoSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_FLAMER]))->mNoAmmoSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_BLASTER]))->mNoAmmoSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_ROCKETLAUNCHER]))->mNoAmmoSoundNumber = iSound;

	iSound = mSDL->loadSound("sounds/ammotake.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	((Weapon *)(mUsableItems[UIT_HANDGUN]))->getAmmo()->mTakeSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_SHOTGUN]))->getAmmo()->mTakeSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_MASHINEGUN]))->getAmmo()->mTakeSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_FLAMER]))->getAmmo()->mTakeSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_BLASTER]))->getAmmo()->mTakeSoundNumber = iSound;
	((Weapon *)(mUsableItems[UIT_ROCKETLAUNCHER]))->getAmmo()->mTakeSoundNumber = iSound;

	iSound = mSDL->loadSound("sounds/ooopstryk.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_HANDGUN]->mTakeSoundNumber = iSound;
	mUsableItems[UIT_SHOTGUN]->mTakeSoundNumber = iSound;
	mUsableItems[UIT_MASHINEGUN]->mTakeSoundNumber = iSound;
	mUsableItems[UIT_FLAMER]->mTakeSoundNumber = iSound;
	mUsableItems[UIT_BLASTER]->mTakeSoundNumber = iSound;
	mUsableItems[UIT_ROCKETLAUNCHER]->mTakeSoundNumber = iSound;

	iSound = mSDL->loadSound("sounds/karta.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_RED_CARD]->mUseSoundNumber = iSound;
	mUsableItems[UIT_GREEN_CARD]->mUseSoundNumber = iSound;
	mUsableItems[UIT_BLUE_CARD]->mUseSoundNumber = iSound;

	iSound = mSDL->loadSound("sounds/ooo.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_RED_CARD]->mTakeSoundNumber = iSound;
	mUsableItems[UIT_GREEN_CARD]->mTakeSoundNumber = iSound;
	mUsableItems[UIT_BLUE_CARD]->mTakeSoundNumber = iSound;
	mSimpleItems[SIT_BOMB]->mTakeSoundNumber = iSound;

	iSound = mSDL->loadSound("sounds/apteczka.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mSimpleItems[SIT_MEDIPACK]->mTakeSoundNumber = iSound;
	mSimpleItems[SIT_POTION]->mTakeSoundNumber = iSound;
	mSimpleItems[SIT_BEER]->mTakeSoundNumber = iSound;

	iSound = mSDL->loadSound("sounds/switch.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mUsableItems[UIT_HAND]->mUseSoundNumber = iSound;

	iSound = mSDL->loadSound("sounds/pik.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mLowEnergySoundNumber = iSound;
	return true;
}

bool CCytadelaGame::loadIGameCtrlSounds()
{
	int32_t iSound = mSDL->loadSound("sounds/arrgh2.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mGameCtrl->mCollisionSounds[CSND_WALL_BUMP] = iSound;

	iSound = mSDL->loadSound("sounds/arrgh.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mGameCtrl->mCollisionSounds[CSND_PLAYER_HIT] = iSound;
	mGameCtrl->mCollisionSounds[CSND_TOUCH_FLAMES] = iSound;

	iSound = mSDL->loadSound("sounds/explode.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mGameCtrl->mCollisionSounds[CSND_EXPLOSION] = iSound;

	mGameCtrl->mDoorSounds[DOORTYPE_UNDEFINED] = -1;
	iSound = mSDL->loadSound("sounds/door1.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mGameCtrl->mDoorSounds[DOORTYPE_VERTICAL_1] = iSound;
	iSound = mSDL->loadSound("sounds/door3.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mGameCtrl->mDoorSounds[DOORTYPE_VERTICAL_2] = iSound;
	iSound = mSDL->loadSound("sounds/door2.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
		return false;
	mGameCtrl->mDoorSounds[DOORTYPE_HORIZONTAL] = iSound;

	return true;
}

void CCytadelaGame::initUserInterface()
{
	//assign images to OSD elements
	//background
	mOGL->mOSD.setImage(OSDIMG_PANEL, OSD_PANEL);
	//hand
	mOGL->mOSD.setImage(OSDIMG_WEAPON_HAND, OSD_WEAPON);
	//bomb
	mOGL->mOSD.setImage(OSDIMG_BOMB_0 + mBomb, OSD_BOMB);
	//plotter
	mOGL->mOSD.setImage(OSDIMG_PLOTTER, OSD_PLOTTER);
	//text
	mOGL->mOSD.setImage(OSDIMG_OSDFONT, OSD_TEXT);

	//draw the counters
	drawEnergyCounter();
	drawAmmoCounter();
	drawCardsCounters();
}

bool CCytadelaGame::runGame(char *szLevel, GameState *pState)
{
	if(not loadImages())
	{
		mLastErrorDesc = "ERROR: COULD NOT LOAD GAME'S RESOURCES!";
		return false;
	}
	if(not loadItemSounds())
	{
		mLastErrorDesc = "ERROR: COULD NOT LOAD GAME'S RESOURCES!";
		return false;
	}
	//create game objects interface and load a file into it
	GOInterface *objects = createGOInterface();
	if(not objects)
	{
		fprintf(stderr, "CCytadelaGame::runGame(): !objects - Out of memory\n");
		mLastErrorDesc = "ERROR: OUT OF MEMORY!";
		return false;
	}
	CMFile file;
	if(not file.open(szLevel))
	{
		fprintf(stderr, "CCytadelaGame::runGame(): !CMFile::open(%s)\n", szLevel);
		mLastErrorDesc = "ERROR: COULD NOT LOAD A LEVEL MAP!";
		objects->release();
		return false;
	}
	if(not file.load(objects))
	{
		fprintf(stderr, "CCytadelaGame::runGame(): !CMFile::load(%s)\n", szLevel);
		mLastErrorDesc = "ERROR: COULD NOT LOAD A LEVEL MAP!";
		objects->release();
		file.close();
		return false;
	}
	//add 3Dg interface to rendering interface
	if(not mOGL->setGOInterface(objects))
	{
		fprintf(stderr, "CCytadelaGame::runGame(): !setGOInterface()\n");
		mLastErrorDesc = "ERROR: COULD NOT LOAD A LEVEL MAP!";
		objects->release();
		file.close();
		return false;
	}
	//initial position
	Point3Dg point;
	if(not file.readInitialPosition(point))
	{
		fprintf(stderr, "CCytadelaGame::runGame(): !CMFile::readInitialPosition()\n");
		mLastErrorDesc = "ERROR: COULD NOT LOAD A LEVEL MAP!";
		objects->release();
		file.close();
		return false;
	}
	mX = point.x;
	mZ = point.z;
	//initial direction
	if(not file.readInitialDirection(point))
	{
		fprintf(stderr, "CCytadelaGame::runGame(): !CMFile::readInitialDirection()\n");
		mLastErrorDesc = "ERROR: COULD NOT LOAD A LEVEL MAP!";
		objects->release();
		file.close();
		return false;
	}
	mRotY = point.y;
	processRotationAngle();
	file.close();

	//create game control interface
	mGameCtrl = createGameCtrlInterface();
	if(not mGameCtrl)
	{
		fprintf(stderr, "CCytadelaGame::runGame(): !mGameCtrl - Out of memory\n");
		mLastErrorDesc = "ERROR: OUT OF MEMORY!";
		objects->release();
		return false;
	}
	//set initial values
	UpdateData sUpdData = {&mX, &mZ, mSin, -mCos, mRotY, mDiff, 0, 0, 0, mPosition};
	if(not mGameCtrl->init(&sUpdData, 35.0f, objects, mOGL, mTexts, mConfig))
	{
		fprintf(stderr, "CCytadelaGame::runGame(): !mGameCtrl->init()\n");
		mLastErrorDesc = "ERROR: COULD NOT INITIALIZE THE GAME!";
		mGameCtrl->release();
		objects->release();
		return false;
	}
	if(not loadIGameCtrlSounds())
	{
		mLastErrorDesc = "ERROR: COULD NOT LOAD GAME'S RESOURCES!";
		mGameCtrl->release();
		objects->release();
		return false;
	}

	mGameCtrl->mHighDificulty = mDifficult;
	mGameCtrl->mVisionNoise = pState->bVisionNoise;
	initUserInterface();

	//beer related variables
	uint32_t uiBeerFrameTime = 0;
	float fDrunkRotY = 0.0f, fDrunkX = 0.0f, fDrunkZ = 0.0f;
	//local time-related variables
	uint32_t dwCurTime, uiDeathTimeDelay = 0;
	int32_t iLowEnergySoundDelay = 1000;

	//set the view
	mOGL->setCamMatrix(mX, mZ, mSin, mCos);
	//fade in
	mOGL->renderFrame();
	fadeIn();
	//"Welcome!" text
	mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_WELCOME));
	//clear the input event queue
	mSDL->clearEventQueue();
	//get the current time and enter the main loop
	mLastTime = mSDL->getTicks();
#if defined(DEBUG)
double fps = 0.0;
double avfps = 0.0;
uint32_t cumtime = 0;
uint32_t frames = 0;
#endif
	while(true)
	{
		//time
		dwCurTime = mSDL->getTicks();
		mDiff = dwCurTime - mLastTime;
		if(mDiff == 0)
			continue;
		mLastTime = dwCurTime;
		if(mEnergy < 10)
		{
			iLowEnergySoundDelay -= mDiff;
			if(iLowEnergySoundDelay <= 0)
			{
				mSDL->playSound(mLowEnergySoundNumber, 5);
				iLowEnergySoundDelay = 1000;
			}
		}
#if defined(DEBUG)
fps = 1000.0/(double)mDiff;
frames++;
cumtime += mDiff;
if(cumtime >= 1500) {
	avfps = 1000.0 * frames / (double)cumtime;
	cumtime = frames = 0;
printf("%f, av=%f\n", fps, avfps);
}
#endif
		//if runnin' on slow machine then simulate at least 10 FPS
		//(to eliminate bugs such as walking through walls, etc.)
		if(mDiff > 100)
			mDiff = 100;
		//handle the input
		if(not processInput()) break;
		if(mPause)
		{
			//mPauseStartTime sometimes can be greater than dwCurTime
			//so we need a signed variable for the difference
			int32_t iPauseTime = dwCurTime - mPauseStartTime;
			if(iPauseTime > 10000)
			{
				mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_PAUSED));
				mPauseStartTime = dwCurTime;
			}
			mOGL->renderFrame();
			mOGL->mOSD.drawOSD(mSin, mCos, mDiff);
			mSDL->swapBuffers();
			continue;
		}
		if(mMap)
		{
			//draw map and OSD (the map is redrawn because on slow systems
			//it may flicker if it is not redrawn, but don't know why)
			mOGL->mOSD.drawMap(mPosition[0], mPosition[1]);
			mOGL->mOSD.drawOSD(mSin, mCos, mDiff);
			mSDL->swapBuffers();
			continue;
		}
		mTotalTime += mDiff;
		//new walk speed dependant on exhaust
		if(not mPotion)
			mWalkSpeed = mStandardWalkSpeed * (1.0f - mGameCtrl->mExhaust / 2.27f);
		//but if potion, the exhust desn't matter
		else
		{
			if(mDiff > mPotion)
				mPotion = 0;
			else
				mPotion -= mDiff;
		}
		//if drunk
		if(mBeer > 0)
		{
			mBeer -= mDiff;
			uiBeerFrameTime += mDiff;
			//chnge the values max 10 times a second
			if(uiBeerFrameTime > 50)
			{
				//random rotation in random direction
				fDrunkRotY = (float)((rand()%3) - 1) / 1000.0f;
				//random movement
				fDrunkX = ((rand()%3) - 1) / 20.0f;
				fDrunkZ = ((rand()%3) - 1) / 20.0f;
				uiBeerFrameTime = 0;
			}
			mRotY += fDrunkRotY * mDiff;
			processRotationAngle();
			mX += fDrunkX * mDiff;
			mZ += fDrunkZ * mDiff;
		}
		//if player went out of level area
		if(mX > 6400.0f) mX -= 6400.0f;
		if(mX < 0.0f) mX += 6400.0f;
		if(mZ < -6400.0f) mZ += 6400.0f;
		if(mZ > 0.0f) mZ -= 6400.0f;
		//update the game control interface
		UpdateData upddata = {&mX, &mZ, mSin, -mCos, mRotY, mDiff, 0, 0, 0, mPosition};
		//if returned 'false' then the level has been finished
		if(not mGameCtrl->update(&upddata))
		{
			//render actualized scene before fade-out so the most recent image
			//will be faded-out
			mOGL->setCamMatrix(mX, mZ, mSin, mCos);
			mOGL->renderFrame();
			fadeOut();
			break;
		}
		//if teleport collision occured then substract the time spent on it
		//from the total gameplay time (actualize the mLastTime)
		mLastTime += upddata.uiTeleportationTime;
		//if equipment collision (iTakenItem == 0 is not an equipment
		//material for sure - the eq mtl nums are in range from 56 to 74)
		if(upddata.iTakenItem)
			takeItem(upddata.iTakenItem);
		//decrease the energy (and finish the game if no energy left, but delay
		//the death for 0.5 second)
		if(not decEnergy(upddata.uiLostEnergy))
		{
			if(uiDeathTimeDelay > 500)
			{
				mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_YOU_ARE_DEAD));
				drawBloodOnScreen();
				break;
			}
			uiDeathTimeDelay += mDiff;
		}
		//draw frame
		mOGL->setCamMatrix(mX, mZ, mSin, mCos);
		mOGL->renderFrame();
		mOGL->mOSD.drawOSD(mSin, mCos, mDiff);
		mSDL->swapBuffers();
	}
	//copy game state
	pState->dwTotalTime = mTotalTime;
	pState->wEnergy = mEnergy;
	pState->wAmmo[GSAM_HANDGUN] = ((Weapon *)mUsableItems[UIT_HANDGUN])->getAmmo()->getAmount();
	pState->wAmmo[GSAM_SHOTGUN] = ((Weapon *)mUsableItems[UIT_SHOTGUN])->getAmmo()->getAmount();
	pState->wAmmo[GSAM_MASHINEGUN] = ((Weapon *)mUsableItems[UIT_MASHINEGUN])->getAmmo()->getAmount();
	pState->wAmmo[GSAM_FLAMER] = ((Weapon *)mUsableItems[UIT_FLAMER])->getAmmo()->getAmount();
	pState->wAmmo[GSAM_BLASTER] = ((Weapon *)mUsableItems[UIT_BLASTER])->getAmmo()->getAmount();
	pState->wAmmo[GSAM_ROCKETLAUNCHER] = ((Weapon *)mUsableItems[UIT_ROCKETLAUNCHER])->getAmmo()->getAmount();
	if(mUsableItems[UIT_HANDGUN]->getStatus() == WPNS_ACTIVE)
		pState->wWeapons |= GSWPN_HANDGUN;
	if(mUsableItems[UIT_SHOTGUN]->getStatus() == WPNS_ACTIVE)
		pState->wWeapons |= GSWPN_SHOTGUN;
	if(mUsableItems[UIT_MASHINEGUN]->getStatus() == WPNS_ACTIVE)
		pState->wWeapons |= GSWPN_MASHINEGUN;
	if(mUsableItems[UIT_FLAMER]->getStatus() == WPNS_ACTIVE)
		pState->wWeapons |= GSWPN_FLAMER;
	if(mUsableItems[UIT_BLASTER]->getStatus() == WPNS_ACTIVE)
		pState->wWeapons |= GSWPN_BLASTER;
	if(mUsableItems[UIT_ROCKETLAUNCHER]->getStatus() == WPNS_ACTIVE)
		pState->wWeapons |= GSWPN_ROCKETLAUNCHER;
	pState->wKilled = mKilled;
	pState->btBomb = mBomb;
	pState->bDifficult = mDifficult;
	pState->bExitGame = not mFinishedLevel;
	pState->bVisionNoise = mGameCtrl->mVisionNoise;
	//free everything
	mGameCtrl->release();
	mOGL->mOSD.deleteImages();
	mSDL->freeSounds();
	objects->release();
	return true;
}

void CCytadelaGame::takeWeapon(Weapon *pWeapon)
{
	pWeapon->setStatus(WPNS_ACTIVE);
	mSDL->playSound(pWeapon->mTakeSoundNumber, 1);
}

void CCytadelaGame::takeCard(Card *pCard)
{
	pCard->mCardsCount++;
	mSDL->playSound(pCard->mTakeSoundNumber, 1);
}

void CCytadelaGame::takeAmmo(Ammo *ammo)
{
	ammo->take();
	mSDL->playSound(ammo->mTakeSoundNumber, 2);
}

void CCytadelaGame::takeItem(int32_t iMtlNum)
{
	switch(iMtlNum)
	{
		case 56:
			takeWeapon((Weapon *)mUsableItems[UIT_HANDGUN]);
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_WEAPON_1));
			return;
		case 57:
			takeWeapon((Weapon *)mUsableItems[UIT_SHOTGUN]);
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_WEAPON_2));
			return;
		case 58:
			takeWeapon((Weapon *)mUsableItems[UIT_MASHINEGUN]);
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_WEAPON_3));
			return;
		case 59:
			takeWeapon((Weapon *)mUsableItems[UIT_FLAMER]);
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_WEAPON_4));
			return;
		case 60:
			takeWeapon((Weapon *)mUsableItems[UIT_BLASTER]);
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_WEAPON_5));
			return;
		case 61:
			takeWeapon((Weapon *)mUsableItems[UIT_ROCKETLAUNCHER]);
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_WEAPON_6));
			return;
		case 62:
			takeAmmo(((Weapon *)mUsableItems[UIT_HANDGUN])->getAmmo());
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_AMMO_1));
			break;
		case 63:
			takeAmmo(((Weapon *)mUsableItems[UIT_SHOTGUN])->getAmmo());
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_AMMO_2));
			break;
		case 64:
			takeAmmo(((Weapon *)mUsableItems[UIT_MASHINEGUN])->getAmmo());
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_AMMO_3));
			break;
		case 65:
			takeAmmo(((Weapon *)mUsableItems[UIT_FLAMER])->getAmmo());
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_AMMO_4));
			break;
		case 66:
			takeAmmo(((Weapon *)mUsableItems[UIT_BLASTER])->getAmmo());
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_AMMO_5));
			break;
		case 67:
			takeAmmo(((Weapon *)mUsableItems[UIT_ROCKETLAUNCHER])->getAmmo());
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_AMMO_6));
			break;
		case 68:
		{
			takeCard((Card *)mUsableItems[UIT_RED_CARD]);
			drawCardsCounters();
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_CARD_1));
		}
		break;
		case 69:
		{
			takeCard((Card *)mUsableItems[UIT_GREEN_CARD]);
			drawCardsCounters();
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_CARD_2));
		}
		break;
		case 70:
		{
			takeCard((Card *)mUsableItems[UIT_BLUE_CARD]);
			drawCardsCounters();
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_CARD_3));
		}
		break;
		case 71:
		{
			if(not mGameCtrl->mImmortal)
			{
				mEnergy += mSimpleItems[SIT_MEDIPACK]->getTakeAmount();
				if(mEnergy > 999)
					mEnergy = 999;
				mSDL->playSound(mSimpleItems[SIT_MEDIPACK]->mTakeSoundNumber, 1);
				drawEnergyCounter();
			}
			mOGL->mOSD.addText(mTexts[MEDIPACK_TEXTS].getText());
		}
		return;
		case 72:
		{
			mWalkSpeed = mStandardWalkSpeed * 1.2f;
			mPotion = mSimpleItems[SIT_POTION]->getTakeAmount();
			mSDL->playSound(mSimpleItems[SIT_POTION]->mTakeSoundNumber, 1);
			mOGL->mOSD.addText(mTexts[POTION_TEXTS].getText());
		}
		return;
		case 73:
		{
			mBeer +=  mSimpleItems[SIT_BEER]->getTakeAmount();
			mSDL->playSound(mSimpleItems[SIT_BEER]->mTakeSoundNumber, 1);
			mOGL->mOSD.addText(mTexts[BEER_TEXTS].getText());
		}
		return;
		case 74:
		{
			mBomb += mSimpleItems[SIT_BOMB]->getTakeAmount();
			mSDL->playSound(mSimpleItems[SIT_BOMB]->mTakeSoundNumber, 1);
			//can happen when the player used a cheat code to get the bomb
			if(mBomb > 6)
				mBomb = 6;
			else
			{
				switch(mBomb)
				{
				case 1:
					mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_BOMB_1));
				break;
				case 2:
					mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_BOMB_2));
				break;
				case 3:
					mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_BOMB_3));
				break;
				case 4:
					mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_BOMB_4));
				break;
				case 5:
					mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_BOMB_5));
				break;
				case 6:
					mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_BOMB_6));
				break;
				}
			}
			mOGL->mOSD.setImage(OSDIMG_BOMB_0 + mBomb, OSD_BOMB);
		}
		return;
	}
	drawAmmoCounter();
}

bool CCytadelaGame::decEnergy(uint16_t wEnergy)
{
	if(wEnergy > mEnergy)
		mEnergy = 0;
	else
		mEnergy -= wEnergy;
	drawEnergyCounter();
	if(not mEnergy)
		return false;
	return true;
}

bool CCytadelaGame::processInput()
{
	InputArray sInpArr;
	memset(&sInpArr, 0, sizeof(sInpArr));
	mSDL->readInputGame(&sInpArr);
	//check actions
	for(uint32_t i = 0; i < 5; ++i)
	{
		if(sInpArr.eKey[i] == SDLK_ESCAPE and not mMap)
			return false;
		if(sInpArr.eKey[i] == SDLK_p and sInpArr.bStatus[i] and not mMap)
		{
			mPause = not mPause;
			mOGL->mOSD.mPause = mPause;
			if(mPause)
			{
				mPauseStartTime = mSDL->getTicks();
				mOGL->mOSD.addText(mTexts[PAUSE_ON_TEXTS].getText());
			}
			else
				mOGL->mOSD.addText(mTexts[PAUSE_OFF_TEXTS].getText());
		}
		//in map we don't check for codes
		if(mMap) continue;

		//process the only pressed keys
		if(not sInpArr.bStatus[i]) continue;

		//increase game speed
		if(sInpArr.eKey[i] == SDLK_KP_PLUS)
		{
			mConfig->setNextGameSpeed();
			float fGameSpeed = mConfig->getGameSpeed();
			mStandardWalkSpeed = 0.1f * fGameSpeed;
			if(mPotion)
				mWalkSpeed = mStandardWalkSpeed * 1.2f;
			//display the current speed
			string strGameSpeed = mLocInterface->getText(LOC_GAME_GAME_SPEED);
			char szSpeed[16];
			sprintf(szSpeed, " %.0f", fGameSpeed * 100.0f);
			strGameSpeed.append(szSpeed);
			strGameSpeed.append("%%");
			mOGL->mOSD.addText((char*)strGameSpeed.c_str());
		}
		//decrease game speed
		if(sInpArr.eKey[i] == SDLK_KP_MINUS)
		{
			mConfig->setPrevGameSpeed();
			float fGameSpeed = mConfig->getGameSpeed();
			mStandardWalkSpeed = 0.1f * fGameSpeed;
			if(mPotion)
				mWalkSpeed = mStandardWalkSpeed * 1.2f;
			//display the current speed
			string strGameSpeed = mLocInterface->getText(LOC_GAME_GAME_SPEED);
			char szSpeed[16];
			sprintf(szSpeed, " %.0f", fGameSpeed * 100.0f);
			strGameSpeed.append(szSpeed);
			strGameSpeed.append("%%");
			mOGL->mOSD.addText((char*)strGameSpeed.c_str());
		}

		//increase mouse speed
		if(sInpArr.eKey[i] == SDLK_RIGHTBRACKET)
		{
			mConfig->setNextMouseSpeed();
			//display the current speed
			string strMouseSpeed = mLocInterface->getText(LOC_GAME_MOUSE_SPEED);
			char szSpeed[16];
			sprintf(szSpeed, " %.1f", mConfig->getMouseSpeed());
			strMouseSpeed.append(szSpeed);
			mOGL->mOSD.addText((char*)strMouseSpeed.c_str());
		}
		//decrease mouse speed
		if(sInpArr.eKey[i] == SDLK_LEFTBRACKET)
		{
			mConfig->setPrevMouseSpeed();
			//display the current speed
			string strMouseSpeed = mLocInterface->getText(LOC_GAME_MOUSE_SPEED);
			char szSpeed[16];
			sprintf(szSpeed, " %.1f", mConfig->getMouseSpeed());
			strMouseSpeed.append(szSpeed);
			mOGL->mOSD.addText((char*)strMouseSpeed.c_str());
		}

		//write key code to the text buffer
		mPlayerText[USERTEXT_LENMINUS1] = (char)sInpArr.eKey[i];
		//check the text and move it afterwards
		if(isCode("hotkiss"))
		{
			//ammo
			((Weapon *)mUsableItems[UIT_HANDGUN])->getAmmo()->setAmount(0x29A);
			((Weapon *)mUsableItems[UIT_SHOTGUN])->getAmmo()->setAmount(0x29A);
			((Weapon *)mUsableItems[UIT_MASHINEGUN])->getAmmo()->setAmount(0x29A);
			((Weapon *)mUsableItems[UIT_FLAMER])->getAmmo()->setAmount(0x29A);
			((Weapon *)mUsableItems[UIT_BLASTER])->getAmmo()->setAmount(0x29A);
			((Weapon *)mUsableItems[UIT_ROCKETLAUNCHER])->getAmmo()->setAmount(0x29A);
			//weapons - damage 'em randomly
			int randnum = rand();
			mUsableItems[UIT_HANDGUN]->setStatus((WeaponState)(1 + (1 & (randnum))));
			mUsableItems[UIT_SHOTGUN]->setStatus((WeaponState)(1 + (1 & (randnum >> 1))));
			mUsableItems[UIT_MASHINEGUN]->setStatus((WeaponState)(1 + (1 & (randnum >> 2))));
			mUsableItems[UIT_FLAMER]->setStatus((WeaponState)(1 + (1 & (randnum >> 3))));
			mUsableItems[UIT_BLASTER]->setStatus((WeaponState)(1 + (1  & (randnum >> 4))));
			mUsableItems[UIT_ROCKETLAUNCHER]->setStatus((WeaponState)(1 + (1 & (randnum >> 5))));
			//cards
			((Card *)mUsableItems[UIT_RED_CARD])->mCardsCount =
				((Card *)mUsableItems[UIT_GREEN_CARD])->mCardsCount =
				((Card *)mUsableItems[UIT_BLUE_CARD])->mCardsCount = 0x6;
			drawAmmoCounter();
			drawCardsCounters();
			for(int32_t j = 0; j < USERTEXT_LEN; ++j)
				mPlayerText[j] = mPlayerText[j + 1];
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_CODE_WEAPONS));
			continue;
		}
		if(isCode("hotkiwi"))
		{
			//god mode
			mGameCtrl->mImmortal = not mGameCtrl->mImmortal;
			for(int32_t j = 0; j < USERTEXT_LEN; ++j)
				mPlayerText[j] = mPlayerText[j + 1];
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_CODE_ENERGY));
			continue;
		}
		if(isCode("kbbmorms"))
		{
			//energy
			mEnergy = 0x29A;
			drawEnergyCounter();
			for(int32_t j = 0; j < USERTEXT_LEN; ++j)
				mPlayerText[j] = mPlayerText[j + 1];
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_CODE_ENERGY));
			continue;
		}
		if(isCode("alibaba"))
		{
			//bomb
			mBomb = 6;
			mOGL->mOSD.setImage(OSDIMG_BOMB_0 + 6, OSD_BOMB);
			for(int32_t j = 0; j < USERTEXT_LEN; ++j)
				mPlayerText[j] = mPlayerText[j + 1];
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_TAKEN_BOMB_6));
			continue;
		}
		if(isCode("idaho"))
		{
			//1 square ahead
			mZ -= mCos * 100.0f;
			mX += mSin * 100.0f;

			if(mX > 6400.0f)
				mX -= 6400.0f;
			if(mX < 0.0f)
				mX += 6400.0f;
			if(mZ < -6400.0f)
				mZ += 6400.0f;
			if(mZ > 0.0f)
				mZ -= 6400.0f;
			uint16_t wPrevOffset = (uint16_t)(mPosition[0] * 8 + mPosition[1] * 512);
			mGameCtrl->teleportCamera(mX, mZ);
			mPosition[0] = (uint8_t)(mX / 100);
			mPosition[1] = (uint8_t)(-mZ / 100);
			uint16_t wCurrentOffset = (uint16_t)(mPosition[0] * 8 + mPosition[1] * 512);
			mGameCtrl->closePrevDoor(wPrevOffset, wCurrentOffset);
			for(int32_t j = 0; j < USERTEXT_LEN; ++j)
				mPlayerText[j] = mPlayerText[j + 1];
			continue;
		}
		if(isCode("lorien"))
		{
			//reveal map
			mGameCtrl->revealMap();

			for(int32_t j = 0; j < USERTEXT_LEN; ++j)
				mPlayerText[j] = mPlayerText[j + 1];
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_CODE_MAP));
			continue;
		}
		if(isCode("kitiara"))
		{
			//next level
			//render actualized scene before fade-out so the most recent image
			//will be faded-out
			mOGL->setCamMatrix(mX, mZ, mSin, mCos);
			mOGL->renderFrame();
			fadeOut();
			mFinishedLevel = true;
			for(int32_t j = 0; j < USERTEXT_LEN; ++j)
				mPlayerText[j] = mPlayerText[j + 1];
			return false;
		}
		for(int32_t j = 0; j < USERTEXT_LEN; ++j)
			mPlayerText[j] = mPlayerText[j + 1];
	}

	//actualize action map
	uint8_t *pKMap = sInpArr.pKeyMap;

	//if level map on display
	if(mMap)
	{
		if(pKMap[SDLK_RETURN] or pKMap[SDLK_KP_ENTER] or pKMap[SDLK_LSHIFT] or
				pKMap[SDLK_RSHIFT] or sInpArr.bRMB or sInpArr.bJBtn)
			mMap = false;
		for(uint32_t i = 0; i < 5; ++i)
		{
			if(sInpArr.bStatus[i] and sInpArr.eKey[i] == SDLK_TAB)
			{
				//if status was left 'true' it would turn on the map again
				//in handleInput()
				sInpArr.bStatus[i] = false;
				mMap = false;
			}
		}
		if(not mMap)
		{
			fadeOut();
			//turn map off so if there was vision noise than it will be rendered
			//to fade-in frame
			mOGL->mOSD.mapOff();
			mOGL->renderFrame();
			mOGL->mOSD.drawOSD(mSin, mCos, 0);
			//now the eventual vision noise is not to be rendered (because it
			//would not be faded than) so turn map on again (but the map won't
			//be displayed because now we draw only pre rendered image without
			//map)
			mOGL->mOSD.mapOn();
			fadeIn();
			mOGL->mOSD.mapOff();
			mLastTime = mSDL->getTicks();
		}
		else return true;
	}

	//fire
	if(pKMap[SDLK_RETURN] or pKMap[SDLK_KP_ENTER] or pKMap[SDLK_LSHIFT] or
				pKMap[SDLK_RSHIFT] or sInpArr.bRMB or sInpArr.bJBtn)
		mActionMap[ACT_FIRE] = true;
	else
	{
		mActionMap[ACT_FIRE] = false;
		mFireOnce = false;
	}
	//forward
	if(pKMap[SDLK_UP] or pKMap[SDLK_KP8] or sInpArr.bLMB or sInpArr.bJoyUp)
		mActionMap[ACT_FORWARD] = true;
	else

		mActionMap[ACT_FORWARD] = false;
	//backward
	if(pKMap[SDLK_DOWN] or pKMap[SDLK_KP5] or sInpArr.bJoyDown)
		mActionMap[ACT_BACKWARD] = true;
	else
		mActionMap[ACT_BACKWARD] = false;
	//turn left
	if(pKMap[SDLK_LEFT] or pKMap[SDLK_KP7] or sInpArr.bJoyLeft)
		mActionMap[ACT_TURN_LEFT] = true;
	else
		mActionMap[ACT_TURN_LEFT] = false;
	//turn right
	if(pKMap[SDLK_RIGHT] or pKMap[SDLK_KP9] or sInpArr.bJoyRight)
		mActionMap[ACT_TURN_RIGHT] = true;
	else
		mActionMap[ACT_TURN_RIGHT] = false;
	//strafe left
	if(pKMap[SDLK_KP4])
		mActionMap[ACT_STRAFE_LEFT] = true;
	else
		mActionMap[ACT_STRAFE_LEFT] = false;
	//strafe right
	if(pKMap[SDLK_KP6])
		mActionMap[ACT_STRAFE_RIGHT] = true;
	else
		mActionMap[ACT_STRAFE_RIGHT] = false;

	//if paused than exit here
	if(mPause)
		return true;

	//do actions
	for(uint32_t i = 0; i < 5; ++i)
		if(sInpArr.bStatus[i])
			handleInput(sInpArr.eKey[i]);
	for(uint32_t i = 0; i < (uint32_t)ACT_NUM_ACTIONS; ++i)
		if(mActionMap[i])
			doActions((Actions)i);

	//mouse rotation (only if mouse moved)
	if(sInpArr.siMouseX)
	{
		mRotY += sInpArr.siMouseX * 0.01f * mConfig->getMouseSpeed();
		processRotationAngle();
	}
	return true;
}

void CCytadelaGame::processRotationAngle()
{
	while(mRotY >= 6.2831853f)
		mRotY -= 6.2831853f;
	while(mRotY < 0.0f)
		mRotY += 6.2831853f;
	//calculate sinus and cosinus of rotation angle
	mSin = sinf(mRotY);
	mCos = cosf(mRotY);
}

bool CCytadelaGame::isCode(const char *szCode)
{
	size_t iCodeLen = strlen(szCode);
	for(size_t i = iCodeLen - 1; i > 0; --i)
	{
		if(szCode[i] != mPlayerText[USERTEXT_LEN - iCodeLen + i])
			return false;
	}
	return true;
}

void CCytadelaGame::drawAmmoCounter()
{
	if(mSelectedItem->getType() == UIT_HAND)
	{
		mOGL->mOSD.setImage(OSDIMG_LDIGIT_NA, OSD_AMMO_O);
		mOGL->mOSD.setImage(OSDIMG_LDIGIT_NA, OSD_AMMO_T);
		mOGL->mOSD.setImage(OSDIMG_LDIGIT_NA, OSD_AMMO_H);
		return;
	}
	uint32_t iCount = mSelectedItem->getAmountForCounter();
	uint32_t iH = iCount / 100;
	uint32_t iT = (iCount - (iH * 100)) / 10;
	uint32_t iO = iCount - (iH * 100) - (iT * 10);
	//digits
	mOGL->mOSD.setImage(OSDIMG_LDIGIT_0 + iO, OSD_AMMO_O);
	mOGL->mOSD.setImage(OSDIMG_LDIGIT_0 + iT, OSD_AMMO_T);
	mOGL->mOSD.setImage(OSDIMG_LDIGIT_0 + iH, OSD_AMMO_H);
}

void CCytadelaGame::drawEnergyCounter()
{
	int32_t iH = mEnergy / 100;
	int32_t iT = (mEnergy - (iH * 100)) / 10;
	int32_t iO = mEnergy - (iH * 100) - (iT * 10);
	if(mEnergy >= 10)
	{
		mOGL->mOSD.setImage(OSDIMG_LDIGIT_0 + iO, OSD_NRG_O);
		mOGL->mOSD.setImage(OSDIMG_LDIGIT_0 + iT, OSD_NRG_T);
		mOGL->mOSD.setImage(OSDIMG_LDIGIT_0 + iH, OSD_NRG_H);
	}
	else
	{
		mOGL->mOSD.setImage(OSDIMG_RDIGIT_0 + iO, OSD_NRG_O);
		mOGL->mOSD.setImage(OSDIMG_RDIGIT_0 + iT, OSD_NRG_T);
		mOGL->mOSD.setImage(OSDIMG_RDIGIT_0 + iH, OSD_NRG_H);
	}
}

void CCytadelaGame::drawCardsCounters()
{
	//red
	uint16_t mCardsCount = ((Card *)mUsableItems[UIT_RED_CARD])->mCardsCount;
	int32_t iH = mCardsCount / 100;
	int32_t iT = (mCardsCount - (iH * 100)) / 10;
	int32_t iO = mCardsCount - (iH * 100) - (iT * 10);
	mOGL->mOSD.setImage(OSDIMG_SDIGIT_0 + iO, OSD_RED_CARD);
	//green
	mCardsCount = ((Card *)mUsableItems[UIT_GREEN_CARD])->mCardsCount;
	iH = mCardsCount / 100;
	iT = (mCardsCount - (iH * 100)) / 10;
	iO = mCardsCount - (iH * 100) - (iT * 10);
	mOGL->mOSD.setImage(OSDIMG_SDIGIT_0 + iO, OSD_GREEN_CARD);
	//blue
	mCardsCount = ((Card *)mUsableItems[UIT_BLUE_CARD])->mCardsCount;
	iH = mCardsCount / 100;
	iT = (mCardsCount - (iH * 100)) / 10;
	iO = mCardsCount - (iH * 100) - (iT * 10);
	mOGL->mOSD.setImage(OSDIMG_SDIGIT_0 + iO, OSD_BLUE_CARD);
}

void CCytadelaGame::doActions(Actions eAction)
{
	switch(eAction)
	{
		case ACT_FIRE:
		{
			if(mFireOnce)
				break;
			fire();
		}
		return;
		//forward
		case ACT_FORWARD:
			mZ -= mCos * mDiff * mWalkSpeed;
			mX += mSin * mDiff * mWalkSpeed;
			return;
		//backward
		case ACT_BACKWARD:
			mZ += mCos * mDiff * mWalkSpeed;
			mX -= mSin * mDiff * mWalkSpeed;
			return;
		//rotations
		case ACT_TURN_LEFT:
		{
			mRotY -= mDiff * 0.001f * mConfig->getGameSpeed();
			processRotationAngle();
		}
		return;
		case ACT_TURN_RIGHT:
		{
			mRotY += mDiff * 0.001f * mConfig->getGameSpeed();
			processRotationAngle();
		}
		return;
		//strafe
		case ACT_STRAFE_LEFT:
			mX -= mCos * mDiff * mWalkSpeed;
			mZ -= mSin * mDiff * mWalkSpeed;
			return;
		case ACT_STRAFE_RIGHT:
			mX += mCos * mDiff * mWalkSpeed;
			mZ += mSin * mDiff * mWalkSpeed;
			return;
		case ACT_NUM_ACTIONS:
			break;
	}
}

void CCytadelaGame::handleInput(SDLKey eKey)
{
	switch(eKey)
	{
	case SDLK_SPACE:
		mGameCtrl->useItem(mUsableItems[UIT_HAND]);
		return;
	//weapons
	case SDLK_F1:
		//if actual weapon is damaged then throw it away
		if(mSelectedItem->getStatus() == WPNS_DAMAGED)
			mSelectedItem->setStatus(WPNS_NA);
		mSelectedItem = mUsableItems[UIT_HAND];
		mOGL->mOSD.setImage(OSDIMG_WEAPON_HAND, OSD_WEAPON);
		break;
	case SDLK_F2:
		{
			//if changing to the same weapon then just break (do nothing)
			if(mSelectedItem->getType() == UIT_HANDGUN)
				break;
			//only if player has the weapon
			if(mUsableItems[UIT_HANDGUN]->getStatus() != WPNS_NA)
			{
				//if previous weapon is damaged then throw it away
				if(mSelectedItem->getStatus() == WPNS_DAMAGED)
					mSelectedItem->setStatus(WPNS_NA);
				mSelectedItem = mUsableItems[UIT_HANDGUN];
				mOGL->mOSD.setImage(OSDIMG_WEAPON_HG, OSD_WEAPON);
			}
		}
		break;
	case SDLK_F3:
		{
			if(mSelectedItem->getType() == UIT_SHOTGUN)
				break;
			if(mUsableItems[UIT_SHOTGUN]->getStatus() != WPNS_NA)
			{
				if(mSelectedItem->getStatus() == WPNS_DAMAGED)
					mSelectedItem->setStatus(WPNS_NA);
				mSelectedItem = mUsableItems[UIT_SHOTGUN];
				mOGL->mOSD.setImage(OSDIMG_WEAPON_SG, OSD_WEAPON);
			}
		}
		break;
	case SDLK_F4:
		{
			if(mSelectedItem->getType() == UIT_MASHINEGUN)
				break;
			if(mUsableItems[UIT_MASHINEGUN]->getStatus() != WPNS_NA)
			{
				if(mSelectedItem->getStatus() == WPNS_DAMAGED)
					mSelectedItem->setStatus(WPNS_NA);
				mSelectedItem = mUsableItems[UIT_MASHINEGUN];
				mOGL->mOSD.setImage(OSDIMG_WEAPON_MG, OSD_WEAPON);
			}
		}
		break;
	case SDLK_F5:
		{
			if(mSelectedItem->getType() == UIT_FLAMER)
				break;
			if(mUsableItems[UIT_FLAMER]->getStatus() != WPNS_NA)
			{
				if(mSelectedItem->getStatus() == WPNS_DAMAGED)
					mSelectedItem->setStatus(WPNS_NA);
				mSelectedItem = mUsableItems[UIT_FLAMER];
				mOGL->mOSD.setImage(OSDIMG_WEAPON_FT, OSD_WEAPON);
			}
		}
		break;
	case SDLK_F6:
		{
			if(mSelectedItem->getType() == UIT_BLASTER)
				break;
			if(mUsableItems[UIT_BLASTER]->getStatus() != WPNS_NA)
			{
				if(mSelectedItem->getStatus() == WPNS_DAMAGED)
					mSelectedItem->setStatus(WPNS_NA);
				mSelectedItem = mUsableItems[UIT_BLASTER];
				mOGL->mOSD.setImage(OSDIMG_WEAPON_BL, OSD_WEAPON);
			}
		}
		break;
	case SDLK_F7:
		{
			if(mSelectedItem->getType() == UIT_ROCKETLAUNCHER)
				break;
			if(mUsableItems[UIT_ROCKETLAUNCHER]->getStatus() != WPNS_NA)
			{
				if(mSelectedItem->getStatus() == WPNS_DAMAGED)
					mSelectedItem->setStatus(WPNS_NA);
				mSelectedItem = mUsableItems[UIT_ROCKETLAUNCHER];
				mOGL->mOSD.setImage(OSDIMG_WEAPON_RL, OSD_WEAPON);
			}
		}
		break;
	case SDLK_F8:
		if(mSelectedItem->getStatus() == WPNS_DAMAGED)
			mSelectedItem->setStatus(WPNS_NA);
		mSelectedItem = mUsableItems[UIT_RED_CARD];
		mOGL->mOSD.setImage(OSDIMG_WEAPON_RC, OSD_WEAPON);
		break;

	case SDLK_F9:
		if(mSelectedItem->getStatus() == WPNS_DAMAGED)
			mSelectedItem->setStatus(WPNS_NA);
		mSelectedItem = mUsableItems[UIT_GREEN_CARD];
		mOGL->mOSD.setImage(OSDIMG_WEAPON_GC, OSD_WEAPON);
		break;
	case SDLK_F10:
		if(mSelectedItem->getStatus() == WPNS_DAMAGED)
			mSelectedItem->setStatus(WPNS_NA);
		mSelectedItem = mUsableItems[UIT_BLUE_CARD];
		mOGL->mOSD.setImage(OSDIMG_WEAPON_BC, OSD_WEAPON);
		break;
	//map
	case SDLK_TAB:
	{
		mMap = true;
		//redraw map in memory
		mGameCtrl->redrawMap();
		//render actualized scene before fade-out so the most recent image
		//will be faded-out
		mOGL->setCamMatrix(mX, mZ, mSin, mCos);
		mOGL->renderFrame();
		mOGL->mOSD.drawOSD(mSin, mCos, 0);
		//turn map on in OSD (we do it before fade-out so vision noise will
		//also fade-out - the vision noise is rendered with the OSD, and the OSD
		//is rendered in every fade-out frame, so if we would not turn on the
		//map before fade-out the vision noise would not fade out)
		mOGL->mOSD.mapOn();
		fadeOut();
		//render the map and fade it in
		mOGL->mOSD.drawMap(mPosition[0], mPosition[1]);
		fadeIn();
		string strPosition = mLocInterface->getText(LOC_GAME_LOCATION);
		char szPos[16];
		sprintf(szPos, ": %d - %d", mPosition[0], mPosition[1]);
		strPosition.append(szPos);
		mOGL->mOSD.addText((char*)strPosition.c_str());
	}
	return;
	case SDLK_z:
		mGameCtrl->mVisionNoise = !mGameCtrl->mVisionNoise;
		if(mGameCtrl->mVisionNoise)
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_VISION_NOISE_ON));
		if(not mGameCtrl->mVisionNoise)
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_VISION_NOISE_OFF));
		return;
	default:
		break;
	}
	drawAmmoCounter();
}

void CCytadelaGame::fire()
{
	if(mSelectedItem->getType() != UIT_MASHINEGUN)
		mFireOnce = true;

	mGameCtrl->useItem(mSelectedItem);
	drawAmmoCounter();
	drawCardsCounters();
}

void CCytadelaGame::drawBloodOnScreen()
{
	uint32_t uiTime = mSDL->getTicks(), uiPastTime = 0, uiDiff = 0, uiFillInterval = 0;
	mOGL->mOSD.freeNoise();
	//there are 320 blood lines and this array stores their lengths
	uint8_t pBloodLines[320];
	memset(pBloodLines, 0, sizeof(pBloodLines));
	//five seconds
	while(uiPastTime < 5000)
	{
		if(uiFillInterval >= 24)
		{
			for(int32_t i = 0; i < 320; ++i)
				pBloodLines[i] += (rand() & 1)*uiFillInterval/12;
			uiFillInterval = 0;
		}
		mOGL->mOSD.addBlood(pBloodLines, 320);
		mOGL->mOSD.drawOSD(mSin, mCos, uiDiff);
		mOGL->mOSD.drawBlood();
		mSDL->swapBuffers();
		uiDiff = mSDL->getTicks() - uiTime - uiPastTime;
		uiPastTime += uiDiff;
		uiFillInterval += uiDiff;
	}
	mOGL->mOSD.freeBlood();
}

void CCytadelaGame::fadeIn()
{
	mOGL->mOSD.drawOSD(mSin, mCos, 0);
	mOGL->preDissolve(0.0f, 0.0f, 0.0f, DT_FADEIN);
	uint32_t dwTime = mSDL->getTicks(), dwPastTime = 0;
	while(dwPastTime < 750)
	{
		float fFadeValue = (float)dwPastTime / 750.0f;
		mOGL->dissolveFrame(DT_FADEIN, fFadeValue);
		mOGL->mOSD.drawOSD(mSin, mCos, 0);
		mSDL->swapBuffers();
		dwPastTime = mSDL->getTicks() - dwTime;
	}
}

void CCytadelaGame::fadeOut()
{
	mOGL->preDissolve(0.0f, 0.0f, 0.0f, DT_FADEOUT);
	uint32_t dwTime = mSDL->getTicks(), dwPastTime = 0;
	while(dwPastTime < 750)
	{
		float fFadeValue = (float)dwPastTime / 750.0f;
		mOGL->dissolveFrame(DT_FADEOUT, fFadeValue);
		mOGL->mOSD.drawOSD(mSin, mCos, 0);
		mSDL->swapBuffers();
		dwPastTime = mSDL->getTicks() - dwTime;
	}
}
